@@ -16,7 +16,8 @@
 #:
 #: This should be the same as the version specified in the :file:`setup.py`
 #: file.
-__version__ = '0.9.1'
+__version__ = '0.9.2'
 
-# make the following name available as part of the public API
+# make the following names available as part of the public API
 from .manager import APIManager
+from .views import ProcessingException
